# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

from uhu.updatehub.api import push_package
from uhu.utils import call

from .hardware import SupportedHardwareManager
from .objects import ObjectsManager


class Package:
    """A package represents a group of objects."""

    def __init__(self, version=None, product=None, dump=None):
        if dump is None:
            self.version = version
            self.product = product
            self.objects = ObjectsManager()
            self.supported_hardware = SupportedHardwareManager()
        else:
            self.version = dump.get('version')  # TODO: validate it
            self.product = dump.get('product')  # TODO: validate it
            self.objects = ObjectsManager(dump=dump)
            self.supported_hardware = SupportedHardwareManager(dump=dump)
        self.uid = None

    def to_metadata(self, callback=None):
        """Serialize package as metadata."""
        metadata = {
            'product': self.product,
            'version': self.version,
        }
        metadata.update(self.supported_hardware.to_metadata())
        metadata.update(self.objects.to_metadata(callback))
        return metadata

    def to_template(self, with_version=True):
        """Serialize package to dump to a file."""
        template = {'product': self.product}
        template['version'] = self.version if with_version else None
        template.update(self.objects.to_template())
        template.update(self.supported_hardware.to_template())
        return template

    def push(self, callback=None):
        """Uploads package to UpdateHub server."""
        call(callback, 'start_objects_load')
        metadata = self.to_metadata(callback)
        call(callback, 'finish_objects_load')
        objects = self.objects.to_upload()
        self.uid = push_package(metadata, objects, callback)
        return self.uid

    def __str__(self):
        return '\n'.join([
            'Product: {}'.format(self.product),
            'Version: {}'.format(self.version),
            str(self.supported_hardware),
            str(self.objects),
        ])



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import shutil
import subprocess


COMPRESSORS = {
    # GZIP format: http://www.gzip.org/zlib/rfc-gzip.html#file-format
    'gzip': {
        'signature': b'\x1f\x8b',
        # pigz does not follow symlinks by default so we use 'readlink' to
        # workaround this limitation. The Yocto Project prefer pigz due its
        # better performance and uses it as a gzip alternative thus we need to
        # support both here.
        'test': 'gzip -t $(readlink -f %s)',
        'cmd': "gzip -l $(readlink -f %s) | tail -n 1 | awk '{ print $2}'"
    },
    # LZO format: http://www.lzop.org/download/lzop-1.03.tar.gz
    'lzop': {
        'signature': b'\x89LZO\x00\r\n\x1a\n',
        'test': 'lzop -t %s',
        'cmd': "lzop -l %s | tail -n 1 | awk '{ print $3}'"
    },
    # XZ format: http://tukaani.org/xz/xz-file-format.txt
    'xz': {
        'signature': b'\xfd7zXZ\x00',
        'test': 'xz -t %s',
        'cmd': "xz -l %s --robot | tail -n 1 | awk '{ print $5}'"
    },
}


MAX_COMPRESSOR_SIGNATURE_SIZE = max(
    [len(compressor['signature']) for compressor in COMPRESSORS.values()])


def get_compressor_format(fn):
    """Returns the compression backend for a given file.

    If file is compressed and we support it, return compression
    format, otherwise return None explicitly.
    """
    with open(fn, 'rb') as fp:
        header = fp.read(MAX_COMPRESSOR_SIGNATURE_SIZE)
    for fmt, compressor in COMPRESSORS.items():
        signature = compressor['signature']
        if signature == header[:len(signature)]:
            return fmt
    return None


def is_compressor_supported(compressor):
    """Checks if compressor utility exists.

    This is necessary so we can get the uncompressed size of an
    object.
    """
    return bool(shutil.which(compressor))


def is_valid_compressed_file(fn, compressor_name):
    """Checks if compressed file is a valid one."""
    compressor = COMPRESSORS.get(compressor_name)
    test_cmd = compressor['test'] % fn
    try:
        result = subprocess.check_call(test_cmd, shell=True)
        if result == 0:
            return True
    except subprocess.CalledProcessError:
        pass  # file is corrupted
    return False


def get_uncompressed_size(fn, compressor_name):
    """Returns uncompressed size of a given compressed file."""
    if compressor_name is None:
        return  # It is not a compressed file
    compressor = COMPRESSORS.get(compressor_name)
    if compressor is None:
        err = '"{}" is not supported'
        raise ValueError(err.format(compressor_name))
    if not is_compressor_supported(compressor_name):
        err = '"{}" is not supported by your system.'
        raise SystemError(err.format(compressor_name))
    if not is_valid_compressed_file(fn, compressor_name):
        err = '"{}" is a bad/corrupted {} file.'
        raise ValueError(err.format(fn, compressor_name))
    cmd = compressor['cmd'] % fn
    size = subprocess.check_output(cmd, shell=True)
    return int(size.decode())


def compression_to_metadata(filename):
    compressor = get_compressor_format(filename)
    size = get_uncompressed_size(filename, compressor)
    if size is None:
        return {}
    return {
        'compressed': True,
        'required-uncompressed-size': size,
    }



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import hashlib
import hmac


class UHV1Signature:  # pylint: disable=too-few-public-methods
    """UpdateHub server Signature V1.

    This signature uses the hmac-sha256 hash algorithm to generate the
    final signature.

    The main point in this signature is that both key and message are
    generated based on the request and user credentials.
    """

    def __init__(self, request, access_id, secret):
        self._request = request
        self._access_id = access_id
        self._secret = secret

    def _hashed_canonical_request(self):
        canonical_request = self._request.canonical()
        return hashlib.sha256(canonical_request.encode()).hexdigest()

    def _signed_headers(self):
        """Generates the signed headers.

        Generates a string listing alphabetically all the headers
        included in the canonical request.

        The returned value will be included in the Authorization
        header so the server may recreates the signature.
        """
        headers = [header.lower() for header in self._request.headers.keys()]
        return ';'.join(sorted(headers))

    def _message(self):
        """Generates the message to be signed.

        Its a string based on the server authentication algorithm name
        and version, the request timestamp in ISO-8601 format and the
        canonical request. Each of these values are separated by a
        line break:

        {auth_name}-{auth-version}
        {request_timestamp}
        {canonical_request}
        """
        timestamp = self._request.date.strftime('%Y%m%dT%H%M%SZ')
        return 'UH-V1\n{timestamp}\n{canonical_request}'.format(
            timestamp=timestamp,
            canonical_request=self._hashed_canonical_request(),
        )

    def _key(self):
        """Generates the key to sign the final message.

        The key is a hexadecimal hash generated with hmac-sha256 using
        base_key as key and the request_date as message:

        key = hex(hmac-256(base_key, message))

        --------
        Base key
        --------

        Base key is the concatenation of the name and version of server
        authentication algorithm plus the user secret:

        {auth_name}-{auth_version}-{user_secret}

        If the authentication algorithm name is UH, the version is
        V1 and the user secret is 123, the base key is:

        base_key = 'UH-V1-123'

        ------------
        Request date
        ------------

        The request date used as message in the hmac algorithm must be
        in the YYYYMMDD format:

        message = '19991231'
        """
        base_key = 'UH-V1-{}'.format(self._secret).encode()
        request_date = self._request.date.strftime('%Y%m%d').encode()
        final_key = hmac.new(base_key, request_date, 'sha256')
        return final_key.hexdigest()

    def _signature_hash(self):
        """Generates the final signature.

        Generates the signature using hmac-sha256 with the already
        computed key and message.
        """
        signature = hmac.new(
            self._key().encode(),
            self._message().encode(),
            'sha256'
        )
        return signature.hexdigest()

    @property
    def signature(self):
        """Creates the value for the Authorization header."""
        header = 'UH-V1 Credential={}, SignedHeaders={}, Signature={}'
        return header.format(
            self._access_id,
            self._signed_headers(),
            self._signature_hash(),
        )



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

from copy import deepcopy

from ._object import Modes


class Object:  # pylint: disable=too-few-public-methods, self-cls-assignment

    def __new__(cls, options):
        opts = deepcopy(options)
        mode = opts.pop('mode')
        cls = Modes.get(mode)
        return cls(opts)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import re


class Options:
    registry = {}

    @classmethod
    def get(cls, name):
        option = cls.registry.get(name)
        if option is None:
            raise ValueError('There is no {} option.'.format(name))
        return option

    @classmethod
    def all(cls):
        return list(cls.registry.values())


class OptionType(type):

    def __init__(cls, classname, bases, methods):
        super().__init__(classname, bases, methods)
        if getattr(cls, 'metadata', None) is not None:
            Options.registry[cls.metadata] = cls


class BaseOption(metaclass=OptionType):
    metadata = None
    help = None
    default = None
    verbose_name = None
    cli = None
    # volatile determines if a option must be not exported or
    # considered when loading an object.
    volatile = False
    # For now, requirements only supports "by value" requirement type
    # and it is implemented with a dict, where the key is the option
    # to check value, and the value is what will be checked against
    # option value.
    requirements = {}
    # symmetric determines if this option can have more than on value,
    # one for each installation set.
    symmetric = True

    choices = None
    min = None
    max = None

    type_name = None

    @classmethod
    def validate(cls, value):
        """Must convert, validate and return a value."""
        raise NotImplementedError

    @classmethod
    def humanize(cls, value):
        return value


class AbsolutePathOption(BaseOption):
    type_name = 'absolute_path'

    @classmethod
    def validate(cls, value):
        """Validates if value is an absolute path."""
        value = str(value)
        result = re.match(r'^/[^\0]*', value)
        if result is None:
            err = '"{}" is not a valid absolut path'
            raise ValueError(err.format(value))
        return value


class BooleanOption(BaseOption):
    type_name = 'boolean'

    @classmethod
    def validate(cls, value):
        if isinstance(value, bool):
            return value
        if isinstance(value, str):
            value = value.strip().lower()
            if value in ['yes', 'y']:
                return True
            if value in ['no', 'n']:
                return False
        raise ValueError('Only y, yes, n and no values are allowed')

    @classmethod
    def humanize(cls, value):
        return 'yes' if value else 'no'


class IntegerOption(BaseOption):
    type_name = 'integer'

    min = None
    max = None

    @classmethod
    def validate(cls, value):
        # This must be checked this way since isintance(True, int) is
        # True in Python.
        # pylint: disable=unidiomatic-typecheck
        if type(value) not in (int, str):
            raise ValueError('Only integers are allowed')
        try:
            value = int(value)
        except ValueError:
            raise ValueError('Only integers are allowed')
        if cls.min is not None and value < cls.min:
            err = '{} is lesser than {}'
            raise ValueError(err.format(value, cls.min))
        if cls.max is not None and value > cls.max:
            err = '{} is greater than {}'
            raise ValueError(err.format(value, cls.max))
        return value


class StringOption(BaseOption):
    type_name = 'string'

    min = None
    max = None
    choices = []

    @classmethod
    def validate(cls, value):
        value = str(value)
        if cls.choices and value not in cls.choices:
            err = '"{}" is not in {}'
            raise ValueError(err.format(value, cls.choices))
        if cls.min is not None and len(value) < cls.min:
            err = '{} length is lesser than {}'
            raise ValueError(err.format(value, cls.min))
        if cls.max is not None and len(value) > cls.max:
            err = '{} length is greater than {}'
            raise ValueError(err.format(value, cls.max))
        return value

    @classmethod
    def humanize(cls, value):
        return '"{}"'.format(value)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import base64
import json
import os

from Cryptodome.Hash import SHA256
from Cryptodome.PublicKey import RSA
from Cryptodome.Signature import PKCS1_v1_5


# Environment variables
CHUNK_SIZE_VAR = 'UHU_CHUNK_SIZE'
GLOBAL_CONFIG_VAR = 'UHU_GLOBAL_CONFIG'
LOCAL_CONFIG_VAR = 'UHU_LOCAL_CONFIG'
SERVER_URL_VAR = 'UHU_SERVER_URL'
ACCESS_ID_VAR = 'UHU_ACCESS_ID'
ACCESS_SECRET_VAR = 'UHU_ACCESS_SECRET'
PRIVATE_KEY_FN = 'UHU_PRIVATE_KEY'
CUSTOM_CA_CERTS_VAR = 'UHU_CUSTOM_CA_CERTS'


# Default values
DEFAULT_CHUNK_SIZE = 1024 * 128  # 128 KiB
DEFAULT_GLOBAL_CONFIG_FILE = os.path.expanduser('~/.config/.uhu')
DEFAULT_LOCAL_CONFIG_FILE = '.uhu'
DEFAULT_SERVER_URL = 'http://0.0.0.0'  # TODO: replace by the right URL


def get_chunk_size():
    return int(os.environ.get(CHUNK_SIZE_VAR, DEFAULT_CHUNK_SIZE))


def get_server_url(path=None):
    url = os.environ.get(SERVER_URL_VAR, DEFAULT_SERVER_URL).strip('/')
    if path is not None:
        url = ''.join((url, path))
    return url


def get_global_config_file():
    if os.path.exists(os.path.expanduser('~/.config/')):
        return os.environ.get(GLOBAL_CONFIG_VAR, DEFAULT_GLOBAL_CONFIG_FILE)
    os.mkdir(os.path.expanduser('~/.config/'))
    return os.environ.get(GLOBAL_CONFIG_VAR, DEFAULT_GLOBAL_CONFIG_FILE)


def get_local_config_file():
    return os.environ.get(LOCAL_CONFIG_VAR, DEFAULT_LOCAL_CONFIG_FILE)


def get_credentials():
    access = os.environ.get(ACCESS_ID_VAR)
    secret = os.environ.get(ACCESS_SECRET_VAR)
    if access and secret:
        return access, secret


def get_custom_ca_certs_file():
    return os.environ.get(CUSTOM_CA_CERTS_VAR, None)


def remove_local_config():
    os.remove(get_local_config_file())


def call(obj, name, *args, **kw):
    func = getattr(obj, name, lambda *args, **kw: None)
    func(*args, **kw)


def indent(value, n_indents, all_lines=False):
    """Indent a multline string to right by n_indents.

    If all_lines is set to True, the first line will also be indented,
    otherwise, first line will be 0 padded. This is so since we can
    attach the generated string in an already indented line.
    """
    lines = value.split('\n')
    padding = n_indents * ' '
    lines = ['{}{}'.format(padding, line).rstrip() for line in lines]
    text = '\n'.join(lines)
    if all_lines:
        return text
    return text.strip()


def list_to_str(title, lst):
    """Given a title and a list prints a named ordered list.

    Ex. title=Menu and lst=[eggs, spam] gives:

        Menu:
            0# eggs
            1# spam
    """
    lines = ['{}:'.format(title)]
    for index, elm in enumerate(lst):
        line = '    {}# {}'.format(index, indent(elm, 4))
        lines.append('')
        lines.append(line)
    return '\n'.join(lines)


def sign_dict(dict_, private_key):
    """Serializes a dict to JSON and sign it using RSA."""
    try:
        with open(private_key) as fp:
            key = RSA.importKey(fp.read())
    except TypeError:
        return
    except (FileNotFoundError, ValueError, IndexError):
        raise ValueError('Invalid private key file.')

    signer = PKCS1_v1_5.new(key)

    # encodes message
    message = SHA256.new(json.dumps(dict_, sort_keys=True).encode())

    # sign
    signature = signer.sign(message)
    return base64.b64encode(signature).decode()



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import math
import sys

from progress.spinner import Spinner
from progress.bar import Bar


class BaseCallback:

    def __init__(self):
        self.uploading = False
        self.max = None

    def object_read(self, n_steps=1):
        """Calls self._object_read for n_steps needed."""
        for _ in range(n_steps):
            self._object_read()

    def _object_read(self):
        if self.uploading:
            self.object_read_upload_callback()
        else:
            self.object_read_load_callback()

    def start_package_upload(self, objects):
        self.uploading = True
        self.max = sum(obj['chunks'] for obj in objects)
        self.start_package_upload_callback()

    def finish_package_upload(self):
        self.uploading = False
        self.finish_package_upload_callback()

    def push_finish(self, uid):  # pylint: disable=no-self-use
        print('Finished! Your package UID is {}'.format(uid))

    def object_read_load_callback(self):
        """Must be called when reading object during loading process."""

    def object_read_upload_callback(self):
        """Must be called when reading object during upload process."""

    def start_package_upload_callback(self):
        """Must be called when starting package upload process."""

    def finish_package_upload_callback(self):
        """Must be called when finished package upload process."""


class TTYCallback(BaseCallback):

    def __init__(self):
        super().__init__()
        self.progress = None

    def start_objects_load(self):
        self.progress = Spinner('Loading objects: ')

    def object_read_load_callback(self):
        self.progress.next()

    def object_read_upload_callback(self):
        self.progress.next()

    def finish_objects_load(self):
        self.progress.finish()
        print('\rLoading objects: ok')
        print('Starting uploading objects...', end='', flush=True)

    def start_package_upload_callback(self):
        suffix = '%(percent)d%% ETA: %(eta)ds'
        self.progress = Bar('Uploading objects:', max=self.max, suffix=suffix)

    def finish_package_upload_callback(self):
        print('\033[1K\rUploading objects: ok')


class NoTTYCallback(BaseCallback):

    def __init__(self):
        super().__init__()
        self.current = 0
        self.coeficient = 5
        self.next_step = 0

    def start_objects_load(self):  # pylint: disable=no-self-use
        print('Loading objects: ', end='', flush=True)

    def object_read_upload_callback(self):
        self.current += 1
        progress = math.floor((self.current / self.max) * 100)
        if progress >= self.next_step:
            steps = progress // self.coeficient
            until = steps * self.coeficient
            for step in range(self.next_step, until, self.coeficient):
                print('{}% '.format(step), end='', flush=True)
            self.next_step = until

    def finish_objects_load(self):  # pylint: disable=no-self-use
        print('ok', flush=True)

    def start_package_upload_callback(self):
        print('Uploading objects:', flush=True)

    def finish_package_upload_callback(self):
        print('100%', flush=True)


def get_callback():
    if sys.stdout.isatty():
        return TTYCallback()
    return NoTTYCallback()


def show_cursor():
    print('\033[?25h', end='')



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import hashlib
import math
import os

from ..utils import call, get_chunk_size

from ._options import Options
from .compression import compression_to_metadata
from .install_condition import InstallCondition
from .validators import validate_options


class Modes:
    registry = {}

    @classmethod
    def get(cls, name):
        obj_class = cls.registry.get(name)
        if obj_class is None:
            raise ValueError('There is no {} object class.'.format(name))
        return obj_class

    @classmethod
    def names(cls):
        return sorted(list(cls.registry))


class ObjectType(type):

    def __init__(cls, classname, bases, methods):
        super().__init__(classname, bases, methods)
        # register class into modes registry
        if getattr(cls, 'mode', None) is not None:
            Modes.registry[cls.mode] = cls

        if cls.allow_compression:
            cls.options.append('compressed')
            cls.options.append('required-uncompressed-size')
        if cls.allow_install_condition:
            cls.options.append('install-condition')
            cls.options.append('install-condition-version')
            cls.options.append('install-condition-pattern-type')
            cls.options.append('install-condition-pattern')
            cls.options.append('install-condition-seek')
            cls.options.append('install-condition-buffer-size')
            cls.string_template.insert(0, ('install-condition', ()))
            cls.string_template.insert(
                1, ('install-condition-pattern-type', ()))
            cls.string_template.insert(2, (
                'install-condition-pattern',
                ('install-condition-seek', 'install-condition-buffer-size')))

        # converts strings to Option classes
        cls.options = [Options.get(opt) for opt in cls.options]
        cls.required_options = [
            Options.get(opt) for opt in cls.required_options]
        cls.string_template = [
            (Options.get(opt), [Options.get(child) for child in children])
            for opt, children in cls.string_template]


class BaseObject(metaclass=ObjectType):
    mode = None
    allow_compression = False
    allow_install_condition = False
    options = []
    required_options = []
    string_template = tuple()
    target_types = None

    @classmethod
    def is_required(cls, option):
        return option in cls.required_options

    def __init__(self, values):
        self._values = validate_options(self, values)
        self.chunk_size = get_chunk_size()
        self.md5 = None

    def to_template(self):
        template = {opt.metadata: value
                    for opt, value in self._values.items()
                    if not opt.volatile}
        template['mode'] = self.mode
        return template

    def to_metadata(self, callback=None):
        self.load(callback)
        metadata = {opt.metadata: value for opt, value in self._values.items()}
        metadata['mode'] = self.mode
        metadata.update(self._metadata_install_condition(metadata))
        metadata.update(self._metadata_compression())
        return metadata

    def _metadata_install_condition(self, metadata):
        if not self.allow_install_condition:
            return {}
        return InstallCondition(metadata).to_metadata()

    def _metadata_compression(self):
        if not self.allow_compression:
            return {}
        return compression_to_metadata(self.filename)

    def to_upload(self):
        return {
            'filename': self['filename'],
            'size': self['size'],
            'sha256sum': self['sha256sum'],
            'md5': self.md5,
            'chunks': len(self)
        }

    @property
    def filename(self):
        """Shortcut to returns object filename option."""
        return self['filename']

    @property
    def size(self):
        """Returns the size of object file."""
        return os.path.getsize(self.filename)

    @property
    def exists(self):
        """Checks if file exsits."""
        return os.path.exists(self.filename)

    def update(self, option, value):
        """Updates a given option value."""
        self[option] = value

    def load(self, callback=None):
        """Reads object to set its size, sha256sum and MD5."""
        sha256sum = hashlib.sha256()
        md5 = hashlib.md5()
        for chunk in self:
            sha256sum.update(chunk)
            md5.update(chunk)
            call(callback, 'object_read')
        self['sha256sum'] = sha256sum.hexdigest()
        self['size'] = self.size
        self.md5 = md5.hexdigest()

    def __setitem__(self, key, value):
        try:
            option = Options.get(key)
        except ValueError:
            raise TypeError('You must provide a registered option')
        try:
            self._values[option] = option.validate(value)
        except ValueError:
            raise TypeError('You must provide a valid value.')
        values = {option.metadata: value
                  for option, value in self._values.items()}
        values[key] = value
        self._values = validate_options(self, values)

    def __getitem__(self, key):
        if not isinstance(key, str):
            raise TypeError('You must provide a option metadata name')
        try:
            option = Options.get(key)
        except ValueError:
            raise TypeError('You must provide a registered option')
        if option not in self.options:
            raise ValueError(
                '{} does not support {}'.format(self.mode, option))
        return self._values.get(option)

    def __len__(self):
        """The size of a object is the number of chunks it has."""
        return math.ceil(self.size/self.chunk_size)

    def __iter__(self):
        """Yields every single chunk."""
        with open(self.filename, 'br') as fp:
            for chunk in iter(lambda: fp.read(self.chunk_size), b''):
                yield chunk

    def __str__(self):
        lines = ['{} [mode: {}]\n'.format(self.filename, self.mode)]
        for option, suboptions in self.string_template:
            value = self[option.metadata]
            if value is None:
                continue
            name = '{}:'.format(option.verbose_name)
            suboptions_value = ''
            if suboptions:
                suboptions_line = []
                for suboption in suboptions:
                    suboption_value = self[suboption.metadata]
                    if suboption_value is None:
                        continue
                    suboptions_line.append('{}: {}'.format(
                        suboption.verbose_name,
                        suboption.humanize(suboption_value)))
                if suboptions_line:
                    suboptions_value = ' [{}]'.format(
                        ', '.join(suboptions_line))
            lines.append('    {:<25}{}{}'.format(
                name, option.humanize(value), suboptions_value))
        return '\n'.join(lines)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

from humanize.filesize import naturalsize

from ._options import (
    AbsolutePathOption, BooleanOption, IntegerOption, StringOption)


class Chip0DevicePath(AbsolutePathOption):
    metadata = 'chip_0_device_path'
    help = 'The device path of Chip 0'
    cli = ['--chip0-device-path']
    verbose_name = 'Chip-0 device path'


class Chip1DevicePath(AbsolutePathOption):
    metadata = 'chip_1_device_path'
    help = 'The device path of Chip 1'
    cli = ['--chip1-device-path']
    verbose_name = 'Chip-1 device path'


class ChunkSizeOption(IntegerOption):
    metadata = 'chunk-size'
    default = 131072
    min = 0
    help = ('The size of the buffers (in bytes) used to read '
            'and write (default is 128KiB)')
    cli = ['--chunk-size']
    verbose_name = 'Chunk size'

    @classmethod
    def humanize(cls, value):
        return naturalsize(value, binary=True)


class CompressedOption(BooleanOption):
    metadata = 'compressed'
    volatile = True


class CountOption(IntegerOption):
    metadata = 'count'
    default = -1
    min = -1
    help = ('How many chunk-size blocks must be copied from the source file '
            'to the target. The default value of -1 means all possible bytes '
            'until the end of the file')
    cli = ['--count']
    verbose_name = 'Count'

    @classmethod
    def humanize(cls, value):
        if value == -1:
            return 'all content'
        return value


class FilenameOption(StringOption):
    metadata = 'filename'
    verbose_name = 'filename'
    min = 1


class FilesystemOption(StringOption):
    metadata = 'filesystem'
    choices = ['btrfs', 'ext2', 'ext3', 'ext4', 'vfat',
               'f2fs', 'jffs2', 'ubifs', 'xfs']
    help = 'Filesystem type that must be used to mount the target-device'
    cli = ['--filesystem', '-fs']
    verbose_name = 'filesystem'

    @classmethod
    def humanize(cls, value):
        return value


class FormatOption(BooleanOption):
    metadata = 'format?'
    default = False
    help = ('Specifies whether the target device (target-device) '
            'should be formatted or not.')
    cli = ['--format']
    verbose_name = 'Format device'


class FormatOptionsOption(StringOption):
    metadata = 'format-options'
    symmetric = False
    help = 'Options to format target-device'
    cli = ['--format-options']
    verbose_name = 'options'
    requirements = {FormatOption: True}


class MountOptionsOption(StringOption):
    metadata = 'mount-options'
    help = 'Options to mount the filesystem in target-device'
    cli = ['--mount-options']
    verbose_name = 'Mount options'


class Padding1koption(BooleanOption):
    metadata = '1k_padding'
    help = 'If 1k-padding should be added in the head'
    cli = ['--padding-1k']
    verbose_name = '1k-padding in the head'


class SearchExponentOption(IntegerOption):
    metadata = 'search_exponent'
    help = 'The search exponent'
    cli = ['--search-exponent']
    verbose_name = 'Search exponent'


class SeekOption(IntegerOption):
    metadata = 'seek'
    default = 0
    min = 0
    help = 'How many chunk-size blocks must be skipped in the target file'
    cli = ['--seek']
    verbose_name = 'seek'


class SHA256Option(StringOption):
    metadata = 'sha256sum'
    volatile = True


class SizeOption(IntegerOption):
    metadata = 'size'
    volatile = True


class SkipOption(IntegerOption):
    metadata = 'skip'
    default = 0
    min = 0
    help = 'How many chunk-size blocks must be skipped in the source file'
    cli = ['--skip']
    verbose_name = 'Skip from source'


class TargetOption(StringOption):
    metadata = 'target'
    symmetric = False
    cli = ['--target', '-t']
    help = 'The target itself'
    verbose_name = 'Target'

    @classmethod
    def humanize(cls, value):
        return value


class TargetTypeOption(StringOption):
    metadata = 'target-type'
    cli = ['--target-type', '-tt']
    help = 'The type of target'
    verbose_name = 'Target type'

    @classmethod
    def get_choices(cls, obj):
        if obj is None:
            return cls.choices
        return obj.target_types


class TargetPathOption(AbsolutePathOption):
    metadata = 'target-path'
    symmetric = False
    cli = ['--target-path', '-tp']
    help = ('Target path on which the object must be '
            'installed to (e.g. /vmlinuz). '
            'This is relative to the mount point.')
    verbose_name = 'Target path'


class TruncateOption(BooleanOption):
    metadata = 'truncate'
    default = False
    help = ('True if the file pointed to by the target-path should be open in '
            'truncate mode (erase content before writing)')
    cli = ['--truncate']
    verbose_name = 'truncate'


class UncompressedSizeOption(IntegerOption):
    metadata = 'required-uncompressed-size'
    volatile = True


class VolumeOption(StringOption):
    metadata = 'volume'
    symmetric = False
    help = 'The volume name'
    cli = ['--volume']
    verbose_name = 'Volume name'

    @classmethod
    def humanize(cls, value):
        return value


class InstallConditionOption(StringOption):
    metadata = 'install-condition'
    choices = ['always', 'content-diverges', 'version-diverges']
    default = 'always'
    help = 'When object should be installed'
    cli = ['--install-condition']
    verbose_name = 'Install condition'

    @classmethod
    def humanize(cls, value):
        if value == 'always':
            return 'always install'
        if value == 'content-diverges':
            return 'install when CONTENT diverges'
        if value == 'version-diverges':
            return 'install when VERSION diverges'


class InstallConditionVersionOption(StringOption):
    metadata = 'install-condition-version'
    volatile = True


class InstallConditionPatternTypeOption(StringOption):
    metadata = 'install-condition-pattern-type'
    choices = ['linux-kernel', 'u-boot', 'regexp']
    help = 'What type of pattern to use to retrive version from object'
    cli = ['--install-condition-pattern-type']
    verbose_name = 'Version pattern type'
    requirements = {InstallConditionOption: 'version-diverges'}


class InstallConditionPatternOption(StringOption):
    metadata = 'install-condition-pattern'
    help = 'Regular expression to use when retriving version from object'
    cli = ['--install-condition-pattern']
    verbose_name = 'Version pattern'
    requirements = {InstallConditionPatternTypeOption: 'regexp'}


class InstallConditionSeekOption(IntegerOption):
    metadata = 'install-condition-seek'
    default = 0
    help = 'Place to start reading object to retrive version'
    cli = ['--install-condition-seek']
    verbose_name = 'seek'
    requirements = {InstallConditionPatternTypeOption: 'regexp'}


class InstallConditionBufferSizeOption(IntegerOption):
    metadata = 'install-condition-buffer-size'
    default = -1
    help = 'How much bytes to read object to retrive version'
    cli = ['--install-condition-buffer-size']
    verbose_name = 'buffer size'
    requirements = {InstallConditionPatternTypeOption: 'regexp'}



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

from prompt_toolkit.validation import ValidationError


class CancelPromptException(Exception):
    """Exception that must be raised when user types Ctrl C."""


class ValidationError(ValidationError):  # pylint: disable=function-redefined
    """Same as prompt_toolkit, but with extra message appended."""

    def __init__(self, document, message):
        message = '{}. Type Ctrl-C to cancel.'.format(message)
        position = len(document.text)
        super().__init__(position, message)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import os

from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.completion import WordCompleter

from ..core.object import Modes


class ObjectFilenameCompleter(Completer):

    DIR = 0
    LINK = 1
    FILE = 2

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = None
        self.directory = None
        self.prefix = None
        self.filenames = None

    def get_completions(self, document, complete_event):
        try:
            self.normalize_value(document)
            self.set_base_directory()
            self.set_base_filename()
            self.set_filenames()
            return self.all_completions()
        except OSError:
            pass

    def normalize_value(self, document):
        self.value = document.text_before_cursor

    def set_base_directory(self):
        dirname = os.path.dirname(self.value)
        if dirname:
            self.directory = os.path.dirname(os.path.join('.', self.value))
        else:
            self.directory = '.'

    def set_base_filename(self):
        self.prefix = os.path.basename(self.value)

    def set_filenames(self):
        try:
            names = sorted([(fn, self._get_sort_key(fn))
                            for fn in os.listdir(self.directory)
                            if fn.startswith(self.prefix)])
        except FileNotFoundError:
            names = []
        self.filenames = sorted(names, key=lambda v: v[1])

    def all_completions(self):
        for filename, file_type in self.filenames:
            yield self.set_completion(filename, file_type)

    def set_completion(self, filename, file_type):
        completion = self._set_completion(file_type, filename)
        kwargs = self._set_completion_kwargs(file_type, filename)
        return Completion(completion, **kwargs)

    def _get_sort_key(self, fn):
        if os.path.isdir(fn):
            return self.DIR
        if os.path.islink(fn):
            return self.LINK
        return self.FILE

    def _set_completion(self, file_type, filename):
        completion = filename[len(self.prefix):]
        if file_type == self.DIR:
            completion += '/'
        return completion

    def _set_completion_kwargs(self, file_type, filename):
        kwargs = {
            'display': filename,
            'start_position': 0,
        }
        if file_type == self.DIR:
            kwargs['display'] = '{}/'.format(filename)
        elif file_type == self.LINK:
            kwargs['display'] = '{} (symbolic link)'.format(filename)
        return kwargs


# pylint: disable=too-few-public-methods
class ObjectModeCompleter(WordCompleter):

    def __init__(self):
        super().__init__(Modes.names())


# pylint: disable=too-few-public-methods
class ObjectOptionValueCompleter(WordCompleter):

    def __init__(self, option):
        completions = sorted(option.choices)
        super().__init__(completions)


# pylint: disable=too-few-public-methods
class ObjectUIDCompleter(WordCompleter):

    def __init__(self, package, installation_set):
        objects = enumerate(package.objects[installation_set])
        completions = ['{}# {}'.format(index, obj.filename)
                       for index, obj in objects]
        super().__init__(sorted(completions))


# pylint: disable=too-few-public-methods
class YesNoCompleter(WordCompleter):

    def __init__(self):
        completions = ['y', 'n']
        super().__init__(completions)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

from setuptools import find_packages, setup

from uhu import get_version

setup(
    version=get_version(),
    packages=find_packages(exclude=['tests*']),
)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import re
import string
import struct
from copy import deepcopy
import libarchive


# Utilities

PRINTABLE = string.printable.encode()
KNOWN_PATTERNS = ['linux-kernel', 'u-boot']
CUSTOM_PATTERN = 'regexp'


def read(fp, seek, type_, buffer_size):
    """Retrives a chunk from file and converts it to a given type."""
    fp.seek(seek)
    try:
        return struct.unpack(type_, fp.read(buffer_size))[0]
    except struct.error:
        return None


def check(phrase, regexp):
    """Checks if a phrase matches a given regexp pattern."""
    results = regexp.findall(phrase)
    if results:
        return results[0].decode()


def find(pattern, iterable):
    """Generic function to find some text in some iterable."""
    regexp = re.compile(pattern)
    phrase = b''
    for chunk in iterable:
        for char in chunk:
            if char in PRINTABLE:
                phrase += bytes([char])
            else:
                result = check(phrase, regexp)
                if result:
                    return result
                phrase = b''
    return check(phrase, regexp)


# Linux Kernel utilities

ARM_Z_IMAGE = 0x016F2818
ARM_U_IMAGE = 0x27051956
X86_BZ_IMAGE = (0xaa55, 1)
X86_Z_IMAGE = (0xaa55, 0)


def is_arm_u_image(fp):
    """Checks if an image is ARM uImage."""
    return read(fp, 0, '>I', 4) == ARM_U_IMAGE


def is_arm_z_image(fp):
    """Checks if an image is ARM zImage."""
    return read(fp, 36, '<I', 4) == ARM_Z_IMAGE


def get_x86_generic_image_info(fp):
    """Generic function to retrive Linux kernel info from x86 images."""
    magic = read(fp, 510, '<H', 2)
    data = read(fp, 529, '<c', 1)
    compression = ord(data) if data is not None else None
    return (magic, compression)


def is_x86_bz_image(fp):
    """Checks if an image is x86 bzImage."""
    return get_x86_generic_image_info(fp) == X86_BZ_IMAGE


def is_x86_z_image(fp):
    """Checks if an image is x86 zImage."""
    return get_x86_generic_image_info(fp) == X86_Z_IMAGE


def get_arm_z_image_version(fp):
    """Returns Linux kernel version of an ARM zImage."""
    # In ARM uImage kernel is compressed within the image. To retrive
    # its version, we need find the compressed kernel, uncompress it,
    # and extract the version from the uncompressed data.

    for header in [
            # Headers taken from:
            # https://github.com/torvalds/linux/blob/master/scripts/extract-vmlinux
            b'\x1f\x8b\x08',  # gzip
            b'\xfd7zXZ\x00',  # xz
            b'CZh',           # bzip2
            b'\x5d\x00\x00',  # lzma
            b'\x89\x4c\x5a',  # lzo
            b'\x02!L\x18',    # lz4
            b'(\xb5/\xfd',    # zstd
    ]:
        fp.seek(0)
        try:
            # This could be improved to avoid to reading all file in memory
            fp.seek(fp.read().index(header))

            with libarchive.stream_reader(
                    fp,
                    format_name='raw',
                    filter_name='all',
                    block_size=512,
            ) as archive:
                data_entry = next(iter(archive))
                iterable = data_entry.get_blocks(512)
                pattern = br'Linux version (\S+).*'
                result = find(pattern, iterable)
                if result:
                    return result
        except ValueError:
            continue
        except libarchive.exception.ArchiveError:
            continue
    return


def get_arm_u_image_version(fp):
    """Returns Linux kernel version of an ARM uImage."""
    fp.seek(32)
    data = fp.read(32).strip(b'\0')
    regexp = re.compile(br'(\d+.?\.[^\s]+)')
    version = check(data, regexp)
    return version


def get_x86_generic_version(fp):
    """Retrives Linux kernel version from x86 images."""
    offset = read(fp, 526, '<H', 2)
    fp.seek(offset + 512)  # 0x200
    version = fp.read(512)
    regexp = re.compile(br'(\d+.?\.[^\s]+)')
    return check(version, regexp)


def get_x86_bz_image_version(fp):
    """Returns Linux kernel version of a x86 bzImage."""
    return get_x86_generic_version(fp)


def get_x86_z_image_version(fp):
    """Returns Linux kernel version of a x86 zImage."""
    return get_x86_generic_version(fp)


# Linux Kernel

def get_kernel_version(fp):
    """Returns Linux kernel object version."""
    result = None
    # ARM uImage
    if is_arm_u_image(fp):
        result = get_arm_u_image_version(fp)
    # ARM zImage
    if is_arm_z_image(fp):
        result = get_arm_z_image_version(fp)
    # x86 bzImage
    if is_x86_bz_image(fp):
        result = get_x86_bz_image_version(fp)
    # x86 zImage
    if is_x86_z_image(fp):
        result = get_x86_z_image_version(fp)
    if result is not None:
        return result
    raise ValueError('Cannot retrive kernel version')


# U-Boot

def get_uboot_version(fp):
    """Returns U-Boot object version."""
    fp.seek(0)
    pattern = br'U-Boot(?: SPL)? (\S+) \(.*\)'
    iterable = iter(lambda: fp.read(30), b'')
    result = find(pattern, iterable)
    if result is not None:
        return result
    raise ValueError('Cannot retrive U-Boot version')


# Arbitrary object

def get_object_version(fp, pattern, seek=0, buffer_size=-1):
    """Returns version of any type of object."""
    fp.seek(seek)
    iterable = iter(lambda: fp.read(buffer_size), b'')
    result = find(pattern, iterable)
    if result is not None:
        return result
    raise ValueError('Cannot retrive object version')


def get_version(fn, type_, **kwargs):
    with open(fn, 'rb') as fp:
        if type_ == 'linux-kernel':
            return get_kernel_version(fp)
        if type_ == 'u-boot':
            return get_uboot_version(fp)
        if type_ == 'regexp':
            kwargs = {k: v for k, v in kwargs.items() if v is not None}
            return get_object_version(fp, **kwargs)


def normalize_install_if_different(values):
    """Converts metadata install-if-different key to install-condition."""
    values = deepcopy(values)
    iid = values.pop('install-if-different', None)

    # Without install-if-different
    if iid is None:
        return values

    # Content diverges
    if iid == 'sha256sum':
        values.update({'install-condition': 'content-diverges'})
        return values

    if not isinstance(iid, dict):
        raise TypeError('Cannot parse install if different.')

    version = iid.get('version')
    if version is None:
        raise ValueError('Missing version in install-if-different property.')

    # Known pattern
    pattern = iid.get('pattern')
    if pattern in KNOWN_PATTERNS:
        values.update({
            'install-condition': 'version-diverges',
            'install-condition-version': version,
            'install-condition-pattern-type': pattern,
        })
        return values

    # Custom pattern
    regexp = pattern.get('regexp')
    if regexp is None:
        raise ValueError('Missing regexp in install-if-different property.')
    # TODO: validate seek and buffer-size
    values.update({
        'install-condition': 'version-diverges',
        'install-condition-version': version,
        'install-condition-pattern-type': 'regexp',
        'install-condition-pattern': regexp,
        'install-condition-seek': pattern.get('seek', 0),
        'install-condition-buffer-size': pattern.get('buffer-size', -1),
    })
    return values


class InstallCondition:  # pylint: disable=too-few-public-methods
    METADATA_KEY = 'install-if-different'

    # Conditions
    ALWAYS = 'always'
    CONTENT_DIVERGES = 'content-diverges'
    VERSION_DIVERGES = 'version-diverges'

    def __init__(self, metadata):
        self.filename = metadata['filename']
        self.condition = metadata.pop('install-condition', None)
        self.metadata = metadata
        self.pattern = None

    def to_metadata(self):
        if self.condition == self.CONTENT_DIVERGES:
            return self._format_metadata('sha256sum')
        if self.condition == self.VERSION_DIVERGES:
            return self._metadata_version_diverges()
        if self.condition == self.ALWAYS:
            return {}
        raise ValueError('Invalid install-condition condition.')

    def _metadata_version_diverges(self):
        self.pattern = self.metadata.pop(
            'install-condition-pattern-type', None)
        if self.pattern in KNOWN_PATTERNS:
            return self._metadata_known_pattern()
        if self.pattern == CUSTOM_PATTERN:
            return self._metadata_custom_pattern()
        raise ValueError('Unknown install-condition pattern type.')

    def _metadata_known_pattern(self):
        return self._format_metadata({
            'pattern': self.pattern,
            'version': get_version(self.filename, self.pattern),
        })

    def _metadata_custom_pattern(self):
        regexp = self.metadata.pop('install-condition-pattern')
        seek = self.metadata.pop('install-condition-seek')
        buffer_size = self.metadata.pop('install-condition-buffer-size')
        version = get_version(
            self.filename, CUSTOM_PATTERN, pattern=regexp.encode(),
            seek=seek, buffer_size=buffer_size)
        return self._format_metadata({
            'version': version,
            'pattern': {
                'buffer-size': buffer_size,
                'regexp': regexp,
                'seek': seek,
            },
        })

    def _format_metadata(self, value):
        return {self.METADATA_KEY: value}
